#!/usr/bin/env python
import matplotlib.pyplot as plt
import numpy as np
import sys
import os
gpu_num = sys.argv[1] #str(params['gpu_num'])
os.environ["CUDA_VISIBLE_DEVICES"]=gpu_num
import tensorflow as tf
import tensorflow_probability as tfp
tfd = tfp.distributions
import tensorflow_addons as tfa
#import tensorflow_graphics as tfg
import time
from lal import GreenwichMeanSiderealTime
from astropy.time import Time
from astropy import coordinates as coord
import corner
import shutil
import h5py
import json
from scipy.spatial.distance import jensenshannon
from scipy.stats import gaussian_kde
from collections import namedtuple
from universal_divergence import estimate
import pandas as pd
import bilby
from bilby.core.prior import Uniform
from tensorflow.keras import mixed_precision
mixed_precision.set_global_policy('float32')

print('... running on GPU {}'.format(gpu_num))

# constants
fourpisq = 4.0*np.pi*np.pi
lntwopi = tf.math.log(2.0*np.pi)
lnfourpi = tf.math.log(4.0*np.pi)

# Let GPU consumption grow as needed
config = tf.compat.v1.ConfigProto()
config.gpu_options.allow_growth = True
session = tf.compat.v1.Session(config=config)
print('... letting GPU consumption grow as needed')

class DataGenerator(tf.keras.utils.Sequence):
    'Generates data for Keras'
    def __init__(self, input_dir, batch_size=512, samples_per_epoch=20000, x_dim=15, y_dim=1024, n_channels=3, test_data=False):
        'Initialization'
        self.x_dim = x_dim
        self.y_dim = y_dim
        self.batch_size = batch_size
        self.fvec = tf.range(params['ndata']/2 + 1)/params['duration'] 
        self.n_channels = n_channels
        self.samples_per_epoch = samples_per_epoch
        if test_data:
            temp = sorted(os.listdir(input_dir),key=lambda x: int(x[:-5].split('_')[1]))
        else:
            temp = os.listdir(input_dir)
        self.filenames = [os.path.join(input_dir,f) for f in temp]     
        self.test_data = test_data
        self.num_files = len(self.filenames)
        x_temp,_ = self.__read_file(self.filenames[0])
        self.samples_per_file = x_temp.shape[0]
        self.files_per_epoch = int(np.ceil(float(self.samples_per_epoch)/float(self.samples_per_file)))
        self.__read_files()
        self.__len__()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(float(self.samples_per_epoch)/float(self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        return self.x[index*self.batch_size:(index+1)*self.batch_size,:], self.y[index*self.batch_size:(index+1)*self.batch_size,:,:]

    def __read_files(self):
        'Read in an epoch of the data from files'
        idx = np.arange(self.num_files)
        if not self.test_data:
            np.random.shuffle(idx)
        self.x, self.y = self.__read_file(self.filenames[idx[0]])
        print('read file {} index'.format(self.filenames[idx[0]],idx[0]))
        for j in idx[1:self.files_per_epoch]:
            temp_x, temp_y = self.__read_file(self.filenames[j])
            self.x = tf.concat([self.x,temp_x],axis=0)
            self.y = tf.concat([self.y,temp_y],axis=0)
            print('read file {} index'.format(self.filenames[j],j))      

    def __read_file(self,filename):
        'Read in data from 1 file'
        print(filename)
        x = np.reshape(np.array(h5py.File(filename, 'r')['x_data']),(-1,self.x_dim))
        if self.test_data:
            y = np.transpose(np.reshape(np.array(h5py.File(filename, 'r')['y_data_noisy']),(-1,self.n_channels,self.y_dim)),[0,2,1])
        else:
            y = np.transpose(np.reshape(np.array(h5py.File(filename, 'r')['y_data_noisefree']),(-1,self.n_channels,self.y_dim)),[0,2,1])
        rand_pars = h5py.File(filename, 'r')['rand_pars'][:]

        # convert ra to hour angle
        x = convert_sky(x, params['ref_geocent_time'], rand_pars)

        # convert psi and phi to reduced parameter space
        for i,k in enumerate(rand_pars):
            if k.decode('utf-8')=='psi':
                psi_idx = i
            if k.decode('utf-8')=='phase':
                 phi_idx = i
    
        # convert phi to X=phi+psi and psi on ranges [0,pi] and [0,pi/2] repsectively - both periodic and in radians
        x[:,psi_idx], x[:,phi_idx] = psiphi_to_psiX(x[:,psi_idx],x[:,phi_idx])
     
        # Normalise the source parameters
        for i,k in enumerate(rand_pars):
            par_min = k.decode('utf-8') + '_min'
            par_max = k.decode('utf-8') + '_max'

            # normalize by bounds - override psi and phase
            if par_min=='psi_min':
                x[:,i]=x[:,i]/(np.pi/2.0)
            elif par_min=='phase_min':
                x[:,i]=x[:,i]/np.pi
            else:
                x[:,i]=(x[:,i] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])

        # extract inference parameters from all source parameters loaded earlier
        idx = []
        infparlist = ''
        for k in params['inf_pars']:
            infparlist = infparlist + k + ', '
            for i,q in enumerate(rand_pars):
                m = q.decode('utf-8')
                if k==m:
                    idx.append(i)
        x = tf.cast(x[:,idx],dtype=tf.float32)
        y = tf.cast(y,dtype=tf.float32)

        if not self.test_data:
            # randomise distance
            old_d = bounds['luminosity_distance_min'] + tf.boolean_mask(x,dist_mask,axis=1)*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
            new_x = tf.random.uniform(shape=tf.shape(old_d), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
            new_d = bounds['luminosity_distance_min'] + new_x*(bounds['luminosity_distance_max'] - bounds['luminosity_distance_min'])
            x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_dist_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_dist_mask),axis=1)
            dist_scale = tf.tile(tf.expand_dims(old_d/new_d,axis=1),(1,tf.shape(y)[1],1))

            # augment the data for this chunk - randomize phase and arrival time
            old_phase = tf.boolean_mask(x,phase_mask,axis=1)*np.pi # this is the X parameters X = psi+phi and has range [0,pi]
            new_x = tf.random.uniform(shape=tf.shape(old_phase), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
            new_phase = new_x*np.pi # this is the X parameters X = psi+phi and has range [0,pi]
            x_data_train = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_phase_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_phase_mask),axis=1)
            old_geocent = bounds['geocent_time_min'] + tf.boolean_mask(x,geocent_mask,axis=1)*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
            new_x = tf.random.uniform(shape=tf.shape(old_geocent), minval=0.0, maxval=1.0, dtype=tf.dtypes.float32)
            new_geocent = bounds['geocent_time_min'] + new_x*(bounds['geocent_time_max'] - bounds['geocent_time_min'])
            x = tf.gather(tf.concat([tf.reshape(tf.boolean_mask(x,not_geocent_mask,axis=1),[-1,tf.shape(x)[1]-1]), tf.reshape(new_x,[-1,1])],axis=1),tf.constant(idx_geocent_mask),axis=1)
            phase_correction = -1.0*tf.complex(tf.cos(new_phase-old_phase),tf.sin(new_phase-old_phase))
            phase_correction = tf.tile(tf.expand_dims(phase_correction,axis=1),(1,3,tf.shape(y)[1]/2 + 1))
            time_correction = -2.0*np.pi*self.fvec*(new_geocent-old_geocent)
            time_correction = tf.complex(tf.cos(time_correction),tf.sin(time_correction))
            time_correction = tf.tile(tf.expand_dims(time_correction,axis=1),(1,3,1))
            y_fft = tf.signal.rfft(tf.transpose(y,[0,2,1]))*phase_correction*time_correction
            y = tf.transpose(tf.signal.irfft(y_fft),[0,2,1])*dist_scale

        return x, y
 

def get_param_index(all_pars,pars,sky_extra=None):
    """ 
    Get the list index of requested source parameter types
    """
    # identify the indices of wrapped and non-wrapped parameters - clunky code
    mask = []
    idx = []

    # loop over inference params
    for i,p in enumerate(all_pars):

        # loop over wrapped params
        flag = False
        for q in pars:
            if p==q:
                flag = True    # if inf params is a wrapped param set flag

        # record the true/false value for this inference param
        if flag==True:
            mask.append(True)
            idx.append(i)
        elif flag==False:
            mask.append(False)

    if sky_extra is not None:
        if sky_extra:
            mask.append(True)
            idx.append(len(all_pars))
        else:
            mask.append(False)

    return mask, idx, np.sum(mask)

def convert_sky(data, geocent_time, pars, single=False):
    """
    Converts right ascension to hour angle or the reverse
    """
    greenwich = coord.EarthLocation.of_site('greenwich')
    t = Time(geocent_time, format='gps', location=greenwich)
    t = t.sidereal_time('mean', 'greenwich').radian

    # compute single instance
    if single:
        return np.remainder(t - data,2.0*np.pi)

    for i,k in enumerate(pars):
        try:
            k = k.decode('utf-8')
        except:
            pass
        if k == 'ra':
            ra_idx = i

    # Check if RA exist
    try:
        ra_idx
    except NameError:
        print('...... RA is fixed. Not converting RA to hour angle or the reverse.')
    else:
        # Iterate over all training samples and convert to hour angle
        for i in range(data.shape[0]):
            data[i,ra_idx] = np.remainder(t - data[i,ra_idx],2.0*np.pi)

    return data

def psiphi_to_psiX(psi,phi):
    """
    Rescales the psi,phi parameters to a different space
    Input and Output in radians
    """
    X = np.remainder(psi+phi,np.pi)
    psi = np.remainder(psi,np.pi/2.0)
    return psi,X

def psiX_to_psiphi(psi,X):
    """
    Rescales the psi,X parameters back to psi, phi
    Input and Output innormalised units [0,1]
    """
    r = np.random.randint(0,2,size=(psi.shape[0],2))
    psi *= np.pi/2.0   # convert to radians 
    X *= np.pi         # convert to radians
    phi = np.remainder(X-psi + np.pi*r[:,0] + r[:,1]*np.pi/2.0,2.0*np.pi).flatten() 
    psi = np.remainder(psi + np.pi*r[:,1]/2.0, np.pi).flatten()
    return psi/np.pi,phi/(np.pi*2.0)  # convert back to normalised [0,1]

def load_samples(params,sampler):
    """
    read in pre-computed posterior samples
    """
    if type("%s" % params['pe_dir']) is str:
        # load generated samples back in
        dataLocations = '%s_%s' % (params['pe_dir'],sampler)
        print('... looking in {} for posterior samples'.format(dataLocations))
    else:
        print('ERROR: input samples directory not a string')
        exit(0)

    # Iterate over requested number of testing samples to use
    for i in range(params['r']):

        filename = '%s/%s_%d.h5py' % (dataLocations,params['bilby_results_label'],i)
        if not os.path.isfile(filename):
            print('... unable to find file {}. Exiting.'.format(filename))
            exit(0)

        print('... Loading test sample -> ' + filename)
        data_temp = {}
        n = 0

        # Retrieve all source parameters to do inference on
        for q in params['bilby_pars']:
            p = q + '_post'
            par_min = q + '_min'
            par_max = q + '_max'
            data_temp[p] = h5py.File(filename, 'r')[p][:]
            if p == 'psi_post':
                print('dealing with psi in samples')
                data_temp[p] = np.remainder(data_temp[p],np.pi)
            if p == 'geocent_time_post':
                data_temp[p] = data_temp[p] - params['ref_geocent_time']
            data_temp[p] = (data_temp[p] - bounds[par_min]) / (bounds[par_max] - bounds[par_min])
            Nsamp = data_temp[p].shape[0]
            n = n + 1
        print('... read in {} samples from {}'.format(Nsamp,filename))

        # place retrieved source parameters in numpy array rather than dictionary
        j = 0
        XS = np.zeros((Nsamp,n))
        for p,d in data_temp.items():
            XS[:,j] = d
            j += 1
        print('... put the samples in an array')

        # Append test sample posteriors to existing array of other test sample posteriors
        rand_idx_posterior = np.random.randint(0,Nsamp,params['n_samples'])
        if i == 0:
            XS_all = np.expand_dims(XS[rand_idx_posterior,:], axis=0)
        else:
            XS_all = np.vstack((XS_all,np.expand_dims(XS[rand_idx_posterior,:], axis=0)))
        print('... appended {} samples to the total'.format(params['n_samples']))
    return XS_all

def plot_losses(train_losses, val_losses, run='testing'):
    """
    plots the losses
    """
    train_losses = np.array(train_losses).reshape(-1,3)
    val_losses = np.array(val_losses).reshape(-1,3)
    epoch = train_losses.shape[0]
    plt.figure(figsize=(14,14))
    plt.semilogx(np.arange(1,epoch+1),train_losses[:,0],'b',label='RECON')
    plt.semilogx(np.arange(1,epoch+1),train_losses[:,1],'r',label='KL')
    plt.semilogx(np.arange(1,epoch+1),train_losses[:,2],'g',label='TOTAL')
    plt.semilogx(np.arange(1,epoch+1),val_losses[:epoch,0],'--b',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_losses[:epoch,1],'--r',alpha=0.5)
    plt.semilogx(np.arange(1,epoch+1),val_losses[:epoch,2],'--g',alpha=0.5)
    plt.xlabel('epoch')
    plt.ylabel('loss')
    plt.legend()
    plt.grid()
    plt.ylim([np.min(1.1*train_losses[int(0.1*epoch):,:]),np.max(1.1*train_losses[int(0.1*epoch):,:])])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/loss.png' % (run))
    if epoch>1:
        plt.xlim([max(1,epoch-100),epoch+1])
        plt.ylim([np.min(train_losses[max(1,epoch-100):,2])-1.0,np.max(train_losses[max(1,epoch-100):,2])+1.0])
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/loss_latest_total.png' % (run))
    plt.close()

    # save loss data to text file
    loss_file = '/data/www.astro/chrism/vitamin_c/{}/loss.txt'.format(run)
    f = open(loss_file,'a+')
    data = tf.reshape(np.concatenate([train_losses[-1,:],val_losses[-1,:]],axis=0),(1,6))
    f.write('%d %f %f %f %f %f %f\n' % (epoch,data[0,0],data[0,1],data[0,2],data[0,3],data[0,4],data[0,5]))
    f.close()

 
def plot_KL(KL_samples, step, run='testing'):
    """
    plots the KL evolution
    """
    # arrives in shape n_kl,n_test,3
    N = KL_samples.shape[0]
    KL_samples = np.transpose(KL_samples,[2,1,0])   # re-order axes
    ls = ['-','--',':']
    c = ['C0','C1','C2','C3']
    fig, axs = plt.subplots(3, sharex=True, sharey=True, figsize=(6.4,14.4))
    for i,kl_s in enumerate(KL_samples):   # loop over samplers
        for j,kl in enumerate(kl_s):     # loop over test cases
            axs[i].semilogx(np.arange(1,N+1)*step,kl,ls[i],color=c[j])
            axs[i].plot(N*step,kl[-1],'.',color=c[j])
            axs[i].grid()
    plt.xlabel('epoch')
    plt.ylabel('KL')
    plt.ylim([-0.2,1.0])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/kl.png' % (run))
    plt.close()

def plot_mode_posterior(samples,x_truth,epoch,idx,run='testing'):
    """
    plots the posteriors in individual latent space modes
    """

    # randonly convert from reduced psi-phi space to full space
    _, psi_idx, _ = get_param_index(params['inf_pars'],['psi'])
    _, phi_idx, _ = get_param_index(params['inf_pars'],['phase'])
    psi_idx = psi_idx[0]
    phi_idx = phi_idx[0]

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=None,
                    levels=[0.90], density=True,
                    plot_density=False, plot_datapoints=False,
                    max_n_ticks=3)

    col = ['tab:blue',
           'tab:orange',
           'tab:green',
           'tab:red',
           'tab:purple',
           'tab:brown',
           'tab:pink',
           'tab:gray',
           'tab:olive',
           'tab:cyan']
 
    par_range = inf_ol_len*[(1e16,-1e16)]
    for i, sample in enumerate(samples):
        sample = sample.numpy()
        # convert back to psi and phi space from X and psi - now on normalised ranges [0,1], and [0,1] respectively 
        sample[:,psi_idx], sample[:,phi_idx] = psiX_to_psiphi(sample[:,psi_idx], sample[:,phi_idx])
        true_x = np.zeros(inf_ol_len)
        true_XS = np.zeros([sample.shape[0],inf_ol_len])
        ol_pars = []
        cnt = 0
        for inf_idx in inf_ol_idx:
            inf_par = params['inf_pars'][inf_idx]
            true_XS[:,cnt] = (sample[:,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
            true_x[cnt] = (x_truth[0,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
            ol_pars.append(inf_par)
            if np.min(true_XS[:,cnt])<par_range[cnt][0]:
                par_range[cnt] = (np.min(true_XS[:,cnt]),par_range[cnt][1])
            if np.max(true_XS[:,cnt])>par_range[cnt][1]:
                par_range[cnt] = (par_range[cnt][0],np.max(true_XS[:,cnt]))
            #par_range[cnt,1] = bounds[inf_par+'_max']
            cnt += 1
        parnames = []
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, ol_pars):
                parnames.append(params['corner_labels'][k])

        # convert to RA
        true_XS = convert_sky(true_XS,params['ref_geocent_time'],ol_pars)
        true_x = convert_sky(np.reshape(true_x,[1,true_XS.shape[1]]),params['ref_geocent_time'],ol_pars).flatten()

        hist_kwargs = dict(density=True,color=col[np.remainder(i,len(col))])
        if i==0:
            try:
                figure = corner.corner(true_XS, **defaults_kwargs,labels=parnames,
                           color=col[np.remainder(i,len(col))], truths=true_x,
                           show_titles=True, range=par_range, hist_kwargs=hist_kwargs)
            except:
                return
        else:
            try:
                corner.corner(true_XS,**defaults_kwargs,
                           color=col[np.remainder(i,len(col))],
                           show_titles=True, range=par_range, fig=figure, hist_kwargs=hist_kwargs)
            except:
                return

    #snr_tot = np.sqrt(np.sum(snrs**2))
    #plt.annotate('SNR = {:.3f} [{:.3f},{:.3f},{:.3f}]'.format(snr_tot,snrs[0],snrs[1],snrs[2]),(0.6,0.95),xycoords='figure fraction',fontsize=18)
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/modes_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

def plot_posterior(samples,x_truth,epoch,idx,run='testing',all_other_samples=None):
    """
    plots the posteriors
    """

    # trim samples from outside the cube
    mask = []
    for s in samples:
        if (np.all(s>=0.0) and np.all(s<=1.0)):
            mask.append(True)
        else:
            mask.append(False)
    samples = tf.boolean_mask(samples,mask,axis=0)
    print('identified {} good samples'.format(samples.shape[0]))
    if samples.shape[0]<100:
        return [-1.0,-1.0,-1.0]

    # randonly convert from reduced psi-phi space to full space
    _, psi_idx, _ = get_param_index(params['inf_pars'],['psi'])
    _, phi_idx, _ = get_param_index(params['inf_pars'],['phase'])
    psi_idx = psi_idx[0]
    phi_idx = phi_idx[0]
    samples = samples.numpy()
    samples[:,psi_idx], samples[:,phi_idx] = psiX_to_psiphi(samples[:,psi_idx], samples[:,phi_idx])

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')
    hist_kwargs_other2 = dict(density=True,color='tab:green')

    ol_pars = []
    for k, inf_idx in enumerate(inf_ol_idx):
        inf_par = params['inf_pars'][inf_idx]
        ol_pars.append(inf_par)
    parnames = []
    for k_idx,k in enumerate(params['rand_pars']):
        if np.isin(k, ol_pars):
            parnames.append(params['corner_labels'][k])

    if all_other_samples is not None:
        n_js = 2*len(all_other_samples) - 1
        js = np.zeros((n_js,len(parnames)))
        for i, other_samples in enumerate(all_other_samples):
            true_post = np.zeros([other_samples.shape[0],bilby_ol_len])
            true_x = np.zeros(inf_ol_len)
            true_XS = np.zeros([samples.shape[0],inf_ol_len])
            cnt = 0
            for inf_idx,bilby_idx in zip(inf_ol_idx,bilby_ol_idx):
                inf_par = params['inf_pars'][inf_idx]
                bilby_par = params['bilby_pars'][bilby_idx]
                true_XS[:,cnt] = (samples[:,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
                true_post[:,cnt] = (other_samples[:,bilby_idx] * (bounds[bilby_par+'_max'] - bounds[bilby_par+'_min'])) + bounds[bilby_par + '_min']
                true_x[cnt] = (x_truth[0,inf_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
                cnt += 1

            # convert to RA
            true_XS = convert_sky(true_XS,params['ref_geocent_time'],ol_pars)
            true_x = convert_sky(np.reshape(true_x,[1,true_XS.shape[1]]),params['ref_geocent_time'],ol_pars).flatten()
            old_true_post = true_post                 

            samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_epoch_{}_event_{}_vit.txt'.format(run,epoch,idx)
            np.savetxt(samples_file,true_XS)

            other_samples_file = '/data/www.astro/chrism/vitamin_c/{}/posterior_samples_event_{}_{}.txt'.format(run,idx,i)
            if not os.path.isfile(other_samples_file):
                np.savetxt(other_samples_file,true_post)

            # calculate js
            g = 0
            for s1,s2 in zip(np.transpose(true_XS),np.transpose(true_post)):
                js[i,g] = calculate_js(s1,s2,ntests=25).median
                g += 1
            if i>0:
                g = 0
                for s1,s2 in zip(np.transpose(last_true_post),np.transpose(true_post)):
                    js[i+len(all_other_samples)-1,g] = calculate_js(s1,s2,ntests=25).median
                    g += 1
            if i==0:
                last_true_post = true_post[:]

            if i==0:
                figure = corner.corner(true_post, **defaults_kwargs,labels=parnames,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
            else:

                corner.corner(true_post,**defaults_kwargs,
                           color='tab:green',
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs_other2)

        #snr_tot = np.sqrt(np.sum(snrs**2))
        #plt.annotate('SNR = {:.3f} [{:.3f},{:.3f},{:.3f}]'.format(snr_tot,snrs[0],snrs[1],snrs[2]),(0.6,0.95),xycoords='figure fraction',fontsize=18)
        corner.corner(true_XS,**defaults_kwargs,
                           color='tab:red',
                           fill_contours=False, truths=true_x,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/comp_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()
        return js, parnames

    else:
        # Get corner parnames to use in plotting labels
        parnames = []
        for k_idx,k in enumerate(params['rand_pars']):
            if np.isin(k, params['inf_pars']):
                parnames.append(params['corner_labels'][k])
        # un-normalise full inference parameters
        full_true_x = np.zeros(len(params['inf_pars']))
        new_samples = np.zeros([samples.shape[0],len(params['inf_pars'])])
        for inf_par_idx,inf_par in enumerate(params['inf_pars']):
            new_samples[:,inf_par_idx] = (samples[:,inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par+'_min']
            full_true_x[inf_par_idx] = (x_truth[0,inf_par_idx] * (bounds[inf_par+'_max'] - bounds[inf_par+'_min'])) + bounds[inf_par + '_min']
        new_samples = convert_sky(new_samples,params['ref_geocent_time'],params['inf_pars'])
        full_true_x = convert_sky(np.reshape(full_true_x,[1,samples.shape[1]]),params['ref_geocent_time'],params['inf_pars']).flatten()       

        figure = corner.corner(new_samples,**defaults_kwargs,labels=parnames,
                           color='tab:red',
                           fill_contours=False, truths=full_true_x,
                           show_titles=True, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/full_posterior_epoch_%d_event_%d.png' % (run,epoch,idx))
        plt.close()

def plot_latent(mu_r1, z_r1, mu_q, z_q, epoch, idx, run='testing'):

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')
    hist_kwargs_other = dict(density=True,color='tab:blue')

    figure = corner.corner(np.array(z_q), **defaults_kwargs,
                           color='tab:blue',
                           show_titles=True, hist_kwargs=hist_kwargs_other)
    corner.corner(np.array(z_r1),**defaults_kwargs,
                           color='tab:red',
                           fill_contours=True,
                           show_titles=True, fig=figure, hist_kwargs=hist_kwargs)
    # Extract the axes
    z_dim = z_r1.shape[1]
    axes = np.array(figure.axes).reshape((z_dim, z_dim))

    # Loop over the histograms
    for yi in range(z_dim):
        for xi in range(yi):
            ax = axes[yi, xi]
            ax.plot(mu_r1[0,:,xi], mu_r1[0,:,yi], "sr")
            ax.plot(mu_q[0,xi], mu_q[0,yi], "sb")
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_epoch_%d_event_%d.png' % (run,epoch,idx))
    plt.close()

def plot_weights(logw, epoch, run='testing'):

    # plot the weights
    plt.figure()
    for lw in logw:
        plt.step(np.arange(params['n_modes']), lw, where='mid')
    plt.xlabel('z dimension')
    plt.ylabel('log(w)')
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_logweight_epoch_%d.png' % (run,epoch))
    plt.close()
    plt.figure()
    maxw = 0.0
    for lw in logw:
        w = np.exp(lw - np.max(lw))
        w = w/np.sum(w)
        if np.max(w)>maxw:
            maxw = np.max(w)
        plt.step(np.arange(params['n_modes']), w, where='mid')
    plt.xlabel('z dimension')
    plt.ylabel('w')
    plt.ylim([0,maxw*1.1])
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/latent_weight_epoch_%d.png' % (run,epoch))
    plt.close()

def plot_r2hist(data, epoch, idx, name='test', run='testing'):

    # define general plotting arguments
    defaults_kwargs = dict(
                    bins=50, smooth=0.9, label_kwargs=dict(fontsize=16),
                    title_kwargs=dict(fontsize=16),
                    truth_color='tab:orange', quantiles=[0.16, 0.84],
                    levels=(0.68,0.90,0.95), density=True,
                    plot_density=False, plot_datapoints=True,
                    max_n_ticks=3)

    # Get corner parnames to use in plotting labels
    parnames = []
    for k_idx,k in enumerate(params['rand_pars']):
        if np.isin(k, params['inf_pars']):
            parnames.append(params['corner_labels'][k])

    # 1-d hist kwargs for normalisation
    hist_kwargs = dict(density=True,color='tab:red')

    try:
        figure = corner.corner(np.array(data) + np.random.normal(loc=0.0,scale=1e-6,size=data.shape), **defaults_kwargs,
                           color='tab:red', #labels=parnames,
                           show_titles=True, hist_kwargs=hist_kwargs)
        plt.savefig('/data/www.astro/chrism/vitamin_c/%s/r2_%s_%d_event_%d.png' % (run,name,epoch,idx))
        plt.close()
    except:
        print('Unable to plot r2 hist data {}'.format(name))
        pass

def calc_median_error(jsvalues, quantiles=(0.16, 0.84)):
    quants_to_compute = np.array([quantiles[0], 0.5, quantiles[1]])
    quants = np.percentile(jsvalues, quants_to_compute * 100)
    summary = namedtuple("summary", ["median", "lower", "upper"])
    summary.median = quants[1]
    summary.plus = quants[2] - summary.median
    summary.minus = summary.median - quants[0]
    return summary


def calculate_js(samplesA, samplesB, ntests=100, xsteps=100):
    js_array = np.zeros(ntests)
    for j in range(ntests):
        nsamples = min([len(samplesA), len(samplesB)])
        A = np.random.choice(samplesA, size=nsamples, replace=False)
        B = np.random.choice(samplesB, size=nsamples, replace=False)
        xmin = np.min([np.min(A), np.min(B)])
        xmax = np.max([np.max(A), np.max(B)])
        x = np.linspace(xmin, xmax, xsteps)
        A_pdf = gaussian_kde(A)(x)
        B_pdf = gaussian_kde(B)(x)

        js_array[j] = np.nan_to_num(np.power(jensenshannon(A_pdf, B_pdf), 2))

    return calc_median_error(js_array)

def plot_js(js_data, epoch, parnames, run='testing'):

    col = ['tab:blue','tab:orange','tab:green','tab:red']
    N = len(parnames)
    fig, ax = plt.subplots()
    for i,js in enumerate(js_data):
        #for x in js:     
        #    ax.semilogy(np.arange(len(parnames)),x,'x',color=col[i])
        boxplot = ax.boxplot(js, showfliers=False, patch_artist=True, whis=[5, 95], widths=0.25, positions=(i - 0.5*(N-1))*0.2 + np.arange(1,N+1))

        for patch in boxplot['boxes']:
            patch.set_facecolor(col[i])

    ax.xaxis.set_ticks(np.arange(len(parnames)))
    ax.set_xticklabels(parnames)
    for tick in ax.get_xticklabels():
        tick.set_rotation(45)
    #ax.set_ylim([])        
    ax.set_yscale('log')
    plt.savefig('/data/www.astro/chrism/vitamin_c/%s/js_%d.png' % (run,epoch), bbox_inches="tight")
    plt.close()

params = '/home/chrism/vitamin_c/params.json'
bounds = '/home/chrism/vitamin_c//bounds.json'
fixed_vals = '/home/chrism/vitamin_c/fixed_vals.json'
#run = time.ctime().replace(' ', '-')
run = time.strftime('%y-%m-%d-%X-%Z')
EPS = 1e-6

path = os.path.join('/data/www.astro/chrism/vitamin_c', run) 
os.mkdir(path) 
shutil.copy('./vitamin_c_functionalAPI.py',path)
shutil.copy('./params.json',path)

# Load parameters files
with open(params, 'r') as fp:
    params = json.load(fp)
with open(bounds, 'r') as fp:
    bounds = json.load(fp)
with open(fixed_vals, 'r') as fp:
    fixed_vals = json.load(fp)

# if doing hour angle, use hour angle bounds on RA
#bounds['ra_min'] = convert_sky(bounds['ra_min'],params['ref_geocent_time'],None,single=True)
#bounds['ra_max'] = convert_sky(bounds['ra_max'],params['ref_geocent_time'],None,single=True)
#print('... converted RA bounds to hour angle')
inf_ol_mask, inf_ol_idx, inf_ol_len = get_param_index(params['inf_pars'],params['bilby_pars'])
bilby_ol_mask, bilby_ol_idx, bilby_ol_len = get_param_index(params['bilby_pars'],params['inf_pars'])

# identify the indices of different sets of physical parameters
vonmise_mask, vonmise_idx_mask, vonmise_len = get_param_index(params['inf_pars'],params['vonmise_pars'])
gauss_mask, gauss_idx_mask, gauss_len = get_param_index(params['inf_pars'],params['gauss_pars'])
sky_mask, sky_idx_mask, sky_len = get_param_index(params['inf_pars'],params['sky_pars'])
ra_mask, ra_idx_mask, ra_len = get_param_index(params['inf_pars'],['ra'])
dec_mask, dec_idx_mask, dec_len = get_param_index(params['inf_pars'],['dec'])
m1_mask, m1_idx_mask, m1_len = get_param_index(params['inf_pars'],['mass_1'])
m2_mask, m2_idx_mask, m2_len = get_param_index(params['inf_pars'],['mass_2'])
psi_mask, psi_idx_mask, psi_len = get_param_index(params['inf_pars'],['psi'])
psiphi_mask, psiphi_idx_mask, psiphi_len = get_param_index(params['inf_pars'],['psi','phase'])
periodic_mask, periodic_idx_mask, periodic_len = get_param_index(params['inf_pars'],['phase','psi','phi_12','phi_jl'])
nonperiodic_mask, nonperiodic_idx_mask, nonperiodic_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','geocent_time','theta_jn','a_1','a_2','tilt_1','tilt_2'])
#idx_mask = np.argsort(gauss_idx_mask + vonmise_idx_mask + m1_idx_mask + m2_idx_mask + sky_idx_mask) # + dist_idx_mask)
idx_mask = np.argsort(m1_idx_mask + m2_idx_mask + gauss_idx_mask + vonmise_idx_mask) # + sky_idx_mask)
dist_mask, dist_idx_mask, dist_len = get_param_index(params['inf_pars'],['luminosity_distance'])
not_dist_mask, not_dist_idx_mask, not_dist_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
phase_mask, phase_idx_mask, phase_len = get_param_index(params['inf_pars'],['phase'])
not_phase_mask, not_phase_idx_mask, not_phase_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','geocent_time','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
geocent_mask, geocent_idx_mask, geocent_len = get_param_index(params['inf_pars'],['geocent_time'])
not_geocent_mask, not_geocent_idx_mask, not_geocent_len = get_param_index(params['inf_pars'],['mass_1','mass_2','luminosity_distance','psi','phase','theta_jn','ra','dec','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
xyz_mask, xyz_idx_mask, xyz_len = get_param_index(params['inf_pars'],['luminosity_distance','ra','dec'])
not_xyz_mask, not_xyz_idx_mask, not_xyz_len = get_param_index(params['inf_pars'],['mass_1','mass_2','psi','phase','geocent_time','theta_jn','a_1','a_2','tilt_1','tilt_2','phi_12','phi_jl'])
idx_xyz_mask = np.argsort(xyz_idx_mask + not_xyz_idx_mask)
idx_dist_mask = np.argsort(not_dist_idx_mask + dist_idx_mask)
idx_phase_mask = np.argsort(not_phase_idx_mask + phase_idx_mask)
idx_geocent_mask = np.argsort(not_geocent_idx_mask + geocent_idx_mask)
idx_periodic_mask = np.argsort(nonperiodic_idx_mask + periodic_idx_mask + ra_idx_mask + dec_idx_mask)
print(xyz_mask)
print(not_xyz_mask)
print(idx_xyz_mask)
#masses_len = m1_len + m2_len
print(params['inf_pars'])
print(vonmise_mask,vonmise_idx_mask)
print(gauss_mask,gauss_idx_mask)
print(m1_mask,m1_idx_mask)
print(m2_mask,m2_idx_mask)
print(sky_mask,sky_idx_mask)
print(idx_mask)

print(nonperiodic_len,periodic_len)


batch_size = 1024
test_size = 8
plot_cadence = 2000

# load the data
training_generator = DataGenerator(params['train_set_dir'], batch_size=batch_size, samples_per_epoch=20000, x_dim=15, y_dim=1024, n_channels=3)
validation_generator = DataGenerator(params['val_set_dir'], batch_size=batch_size, samples_per_epoch=batch_size, x_dim=15, y_dim=1024, n_channels=3)
test_generator = DataGenerator(params['test_set_dir'], batch_size=1, samples_per_epoch=test_size, x_dim=15, y_dim=1024, n_channels=3, test_data=True)
pp_generator = DataGenerator(params['pp_set_dir'], batch_size=1, samples_per_epoch=100, x_dim=15, y_dim=1024, n_channels=3, test_data=True)

#x_data_test, y_data_test_noisefree, y_data_test, snrs_test = load_data(params,bounds,fixed_vals,params['test_set_dir'],params['inf_pars'],max_samples=params['r'],test_data=True)
#x_data_pp, y_data_pp_noisefree, y_data_pp, snrs_pp = load_data(params,bounds,fixed_vals,params['pp_set_dir'],params['inf_pars'],max_samples=100,test_data=True)

# load precomputed samples
bilby_samples = np.array([load_samples(params,'dynesty1'),load_samples(params,'cpnest1')])

#test_dataset = (tf.data.Dataset.from_tensor_slices((x_data_test,y_data_test,snrs_test)).batch(1))
#pp_dataset = (tf.data.Dataset.from_tensor_slices((x_data_pp,y_data_pp,snrs_pp)).batch(1))

class RampCallback(tf.keras.callbacks.Callback):
    def __init__(self, start=100, ramp_length=100, n_cycles=1, ramp_min=0.0, ramp_max=1.0):
        super(RampCallback, self).__init__()
        self.start = start
        self.ramp_length = ramp_length
        self.n_cycles = n_cycles
        self.ramp_min = ramp_min
        self.ramp_max = ramp_max
    def on_train_begin(self, logs=None):
        tf.keras.backend.set_value(self.model.ramp, tf.constant(0.0))
    def on_epoch_begin(self, epoch, logs=None):
        i = (tf.cast(epoch,dtype=tf.float32)-self.start)/(2.0*self.ramp_length)
        if i<0:
            ramp = self.ramp_min
        elif i>=self.n_cycles:
            ramp = self.ramp_max
        else:
            ramp = min(self.ramp_max,self.ramp_min + (self.ramp_max-self.ramp_min)*2.0*np.remainder(i,1.0)) 
        tf.keras.backend.set_value(self.model.ramp, tf.constant(ramp))

class TestCallback(tf.keras.callbacks.Callback):

    def __init__(self, test_generator, pp_generator, bilby_samples, run, n_samples=10000, test_cadence=1000):
        super(TestCallback, self).__init__()
        self.plot_latent = plot_latent
        self.plot_posterior = plot_posterior
        self.plot_mode_posterior = plot_mode_posterior
        self.plot_weights = plot_weights
        self.plot_js = plot_js
        self.test_generator = test_generator
        self.pp_generator = pp_generator
        self.bilby_samples = bilby_samples
        self.test_cadence = test_cadence
        self.n_samples = n_samples
        self.run = run

    def on_epoch_end(self, epoch, logs = None):
        
        # perform test every X epochs
        if (epoch > 0) and (epoch % self.test_cadence == 0):
            logw = []
            all_js = []

            # loop over the test data
            for step,[x,y] in enumerate(self.test_generator):

                # generate the latent space distributions and plot             
                mu_r1, logw_r1, z_r1, mu_q, z_q = self.model.gen_z_samples(x, y, nsamples=self.n_samples)
                logw.append(logw_r1[0,:])
                self.plot_latent(mu_r1,z_r1,mu_q,z_q,epoch,step,run=self.run)
                plt.close('all')               

                # generate the test samples and plot
                start_time_test = time.time()
                x_samples, mu_r2, logvar_r2 = self.model.gen_samples(y, nsamples=self.n_samples)
                end_time_test = time.time()
                print('Epoch: {}, Testing time elapsed for {} samples: {}'.format(epoch+1,self.n_samples,end_time_test - start_time_test))
                js, parnames = self.plot_posterior(x_samples,x,epoch,step,all_other_samples=self.bilby_samples[:,step,:],run=self.run)
                all_js.append(js)
                self.plot_posterior(x_samples,x,epoch,step,run=self.run)
                plt.close('all')                

                # output distribution means and logvars
                plot_r2hist(tf.squeeze(mu_r2),epoch,step,name='mu',run=run)
                plot_r2hist(tf.squeeze(logvar_r2),epoch,step,name='logvar',run=run)
                plt.close('all')

                # generate the plots for the top 4 latent space modes
                mode_samples = self.model.gen_mode_samples(y, nsamples=self.n_samples)
                self.plot_mode_posterior(mode_samples,x,epoch,step,run=self.run)
                plt.close('all')

            # plot ensemble weights and js plots
            self.plot_weights(logw,epoch,run=self.run)
            all_js = np.transpose(np.reshape(all_js,(step+1,2*self.bilby_samples.shape[0]-1,len(parnames))),(1,0,2))
            self.plot_js(all_js,epoch,parnames,run=self.run)
            plt.close('all')

            # make the pp-plots
            results = []
            for step, [x,y] in enumerate(self.pp_generator):
                posterior = dict()
                injections = dict()
                priors = dict()
                samples, _, _ = self.model.gen_samples(y,nsamples=self.n_samples)
                x_batch_pp = x.numpy().astype('float64')
                for i,x in enumerate(x_batch_pp[0]):
                    posterior[str(i)] = samples[:,i]
                    injections[str(i)] = x
                    priors[str(i)] = Uniform(-100,100)
                posterior = pd.DataFrame(dict(posterior))
                result = bilby.result.Result(
                    label="test",
                    injection_parameters=injections,
                    posterior=posterior,
                    search_parameter_keys=injections.keys(),
                    priors=priors)
                results.append(result)
            bilby.result.make_pp_plot(results, filename='/data/www.astro/chrism/vitamin_c/%s/pp_plot_epoch_%d.png' % (self.run,epoch),
                              confidence_interval=[0.68,0.95,0.997])
            plt.close('all')

            # also save the model
            self.model.save_weights('/data/www.astro/chrism/vitamin_c/%s/model.h5' % (self.run))


class LossHistory(tf.keras.callbacks.Callback):
    def __init__(self, run='testing'):
        super(LossHistory, self).__init__()
        self.run = run

    def on_train_begin(self, logs={}):
        self.train_losses = []
        self.val_losses = []

    def on_epoch_end(self, epoch, logs={}):
        self.train_losses.append([logs.get('r_loss'),logs.get('kl_loss'),logs.get('loss')])
        self.val_losses.append([logs.get('val_r_loss'),logs.get('val_kl_loss'),logs.get('val_loss')])
        if np.remainder(epoch+1,5)==0:
            plot_losses(self.train_losses, self.val_losses, run=self.run)
        plt.close('all')

class ReloadCallback(tf.keras.callbacks.Callback):
    def __init__(self, training_generator, reload_cadence=4):
        super(ReloadCallback, self).__init__()
        self.training_generator = training_generator
        self.reload_cadence = reload_cadence

    def on_epoch_begin(self, epoch, logs={}):
        if np.remainder(epoch,self.reload_cadence)==0 and epoch>0:
            print('---> Refreshing training data')
            self.training_generator._DataGenerator__read_files()

#def scheduler(epoch, lr):
#    if np.remainder(epoch+1,1000)==0:
#        return lr * 0.99
#    return lr

class CVAE(tf.keras.Model):
    """Convolutional variational autoencoder."""

    def __init__(self, x_dim_np, x_dim_p, y_dim, n_channels, z_dim, n_modes, run):
        super(CVAE, self).__init__()
        self.z_dim = z_dim
        self.n_modes = n_modes
        self.x_dim_np = x_dim_np # non-periodic
        self.x_dim_p = x_dim_p   # periodic
        self.x_dim = self.x_dim_np + self.x_dim_p + 2 # extra 2 for sky
        self.y_dim = y_dim
        self.corr_norm = 1.0/np.sqrt(y_dim)
        self.n_channels = n_channels #int(n_channels*(n_channels+1)/2)
        self.act = 'relu'
        self.run = run
        self.ramp = tf.Variable(1.0, trainable=False, name='ramp', dtype=tf.float32)
        mom = 0.9

        # the r1 network
        all_input_y = tf.keras.Input(shape=(self.y_dim, self.n_channels))
        #input_y, corr_y = tf.split(all_input_y, num_or_size_splits=2, axis=2)

        #r1c = tf.keras.layers.Reshape((self.y_dim,self.n_channels,1))(all_input_y)   # treat as an image with 1 channel
        #r1c = tf.keras.layers.Permute((2,1,3))(r1c)  # put the detector axis 1st (after batch)
        #conv_layer = tf.keras.layers.Conv1D(filters=128, kernel_size=5, strides=1, activation=self.act)       
        #r1c = tf.keras.layers.TimeDistributed(conv_layer)(r1c)
        #conv_layer = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=self.act)
        #r1c = tf.keras.layers.TimeDistributed(conv_layer)(r1c)
        #conv_layer = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=self.act)
        #r1c = tf.keras.layers.TimeDistributed(conv_layer)(r1c)
        #r1c = tf.keras.layers.Permute((2,1,3))(r1c)  # put the detectors last again
        #r1c = tf.keras.layers.Reshape((-1,self.n_channels*128))(r1c)   # group the detectors and output channels together
        
        #r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=32, strides=1, activation=self.act)(all_input_y)
        #r1c = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=self.act)(r1c)
        #r1c = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=self.act)(r1c)
        #r1c = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, activation=self.act)(r1c)
        #r1c = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, activation=self.act)(r1c)
        #r1c = tf.keras.layers.Flatten()(r1c)

        #r1c0 = tf.keras.layers.Conv1D(filters=64, kernel_size=8, strides=1, padding='same', activation=self.act)(all_input_y)

        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c0)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Add()([r1c,r1c0])
        #r1c0 = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)

        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c0)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Add()([r1c,r1c0])       
        #r1c0 = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)        

        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c0)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Add()([r1c,r1c0])
        #r1c0 = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)

        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c0)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c)
        #r1c = tf.keras.layers.Add()([r1c,r1c0])
        #r1c0 = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)

        #r1c = tf.keras.layers.Conv1D(filters=64, kernel_size=8, strides=1, padding='same', activation=self.act)(input_y)        
        #r1c = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1c)
        #r1c = tf.keras.layers.Conv1D(filters=16, kernel_size=1, strides=1, padding='same', activation=self.act)(r1c)

        #r1cc = tf.keras.layers.Conv1D(filters=64, kernel_size=8, strides=1, padding='same', activation=self.act)(corr_y)
        #r1cc = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(r1cc)
        #r1cc = tf.keras.layers.Conv1D(filters=16, kernel_size=1, strides=1, padding='same', activation=self.act)(r1cc)       
        #r1c = tf.keras.layers.concatenate([r1c,r1cc],axis=2)

        r1c = self.test(all_input_y)
        r1c = tf.keras.layers.Flatten()(r1c)

        r1 = tf.keras.layers.Dense(1024,activation=self.act)(r1c)
        r1 = tf.keras.layers.LayerNormalization()(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=mom)(r1)
        r1 = tf.keras.layers.Dense(1024,activation=self.act)(r1)
        r1 = tf.keras.layers.LayerNormalization()(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=mom)(r1)
        r1 = tf.keras.layers.Dense(1024,activation=self.act)(r1)
        r1 = tf.keras.layers.LayerNormalization()(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=mom)(r1)
        #r1 = tf.keras.layers.Dense(256,activation=self.act)(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=0.5)(r1)
        #r1 = tf.keras.layers.Dense(256,activation=self.act)(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=0.5)(r1)
        #r1 = tf.keras.layers.Dense(256,activation=self.act)(r1)
        #r1 = tf.keras.layers.BatchNormalization(momentum=0.5)(r1)
        r1mu = tf.keras.layers.Dense(self.z_dim*self.n_modes,use_bias=True,bias_initializer=tf.keras.initializers.RandomNormal(mean=0.0, stddev=1.0))(r1c)
        r1logvar = tf.keras.layers.Dense(self.z_dim*self.n_modes,use_bias=True,bias_initializer=tf.keras.initializers.Constant(value=-1.0))(r1c) #tf.keras.initializers.RandomUniform(minval=-4.0, maxval=0.0))(r1)
        r1w = tf.keras.layers.Dense(self.n_modes,use_bias=True,bias_initializer='Zeros')(r1c)
        r1 = tf.keras.layers.concatenate([r1mu,r1logvar,r1w])
        self.encoder_r1 = tf.keras.Model(inputs=all_input_y, outputs=r1)
        print(self.encoder_r1.summary())

        # the q encoder network
        q_input_x = tf.keras.Input(shape=(self.x_dim))
        qx = tf.keras.layers.Flatten()(q_input_x)
        qc = self.test(all_input_y)
        qc = tf.keras.layers.Flatten()(qc) 
        q = tf.keras.layers.concatenate([qc,qx],axis=1)
        #q = tf.keras.layers.Dense(1000,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=mom)(q)
        #q = tf.keras.layers.Dense(2048,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=mom)(q)
        #q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=mom)(q)
        #q = tf.keras.layers.Dense(256,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=0.5)(q)
        #q = tf.keras.layers.Dense(256,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=0.5)(q)
        #q = tf.keras.layers.Dense(256,activation=self.act)(q)
        #q = tf.keras.layers.BatchNormalization(momentum=0.5)(q)
        q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        q = tf.keras.layers.LayerNormalization()(q)
        q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        q = tf.keras.layers.LayerNormalization()(q)
        q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        q = tf.keras.layers.LayerNormalization()(q)
        #q = tf.keras.layers.Dense(1024,activation=self.act)(q)
        q = tf.keras.layers.Dense(2*self.z_dim)(q)
        self.encoder_q = tf.keras.Model(inputs=[all_input_y, q_input_x], outputs=q)
        print(self.encoder_q.summary())

        # the r2 decoder network
        r2_input_z = tf.keras.Input(shape=(self.z_dim))
        r2z = tf.keras.layers.Flatten()(r2_input_z)
        r2c = self.test(all_input_y)
        r2c = tf.keras.layers.Flatten()(r2c)
        r2 = tf.keras.layers.concatenate([r2c,r2z],axis=1)
        #r2 = tf.keras.layers.Dense(1000,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=mom)(r2)
        #r2 = tf.keras.layers.Dense(2048,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=mom)(r2)
        #r2 = tf.keras.layers.Dense(1024,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=mom)(r2)
        #r2 = tf.keras.layers.Dense(256,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=0.5)(r2)
        #r2 = tf.keras.layers.Dense(256,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=0.5)(r2)
        #r2 = tf.keras.layers.Dense(256,activation=self.act)(r2)
        #r2 = tf.keras.layers.BatchNormalization(momentum=0.5)(r2)
        r2 = tf.keras.layers.Dense(1024,activation=self.act)(r2)
        r2 = tf.keras.layers.LayerNormalization()(r2)
        r2 = tf.keras.layers.Dense(1024,activation=self.act)(r2)
        r2 = tf.keras.layers.LayerNormalization()(r2)
        r2 = tf.keras.layers.Dense(1024,activation=self.act)(r2)
        r2 = tf.keras.layers.LayerNormalization()(r2)
        r2mu_np = tf.keras.layers.Dense(self.x_dim_np,activation='sigmoid')(r2)
        r2logvar_np = tf.keras.layers.Dense(self.x_dim_np,use_bias=True)(r2) #,activation=self.act,bias_initializer=tf.keras.initializers.Constant(value=2.0))(r2)
        r2mu_p = tf.keras.layers.Dense(2*self.x_dim_p,use_bias=True)(r2)
        r2logvar_p = tf.keras.layers.Dense(self.x_dim_p,use_bias=True)(r2) #,activation=self.act,bias_initializer=tf.keras.initializers.Constant(value=0.0))(r2)
        r2mu_s = tf.keras.layers.Dense(3,use_bias=True)(r2)
        r2logvar_s = tf.keras.layers.Dense(1,use_bias=True)(r2)
        r2 = tf.keras.layers.concatenate([r2mu_np,r2mu_p,r2mu_s,r2logvar_np,r2logvar_p,r2logvar_s])
        self.decoder_r2 = tf.keras.Model(inputs=[all_input_y, r2_input_z], outputs=r2)
        print(self.decoder_r2.summary())

    def test(self,input_y):

        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(input_y)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(stem)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        #stem = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(stem)  
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(stem)
        stem = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(stem)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(stem)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        #stem = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(stem)
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(stem)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='valid')(stem)
        #stem = tf.keras.layers.LayerNormalization()(stem)
        stem = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(stem)
        # outsize = 256 x 32 = 8192

        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(stem)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(bulk)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(bulk)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(bulk)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(bulk)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.Conv1D(filters=64, kernel_size=3, strides=1, padding='valid')(bulk)
        #bulk = tf.keras.layers.LayerNormalization()(bulk)
        bulk = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(bulk)
        # outsize = 128 x 64 = 8192

        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(bulk)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(tail)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(tail)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(tail)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(tail)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='valid')(tail)
        #tail = tf.keras.layers.LayerNormalization()(tail)
        tail = tf.keras.layers.MaxPool1D(pool_size=2, strides=2)(tail)
        # outsize = 64 x 128 = 8192

        tail = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='valid')(tail)
        # outsize = 64 x 64 = 4096

        return tail

    def IRN(self,input_y,extra=None,batchnorm=True):

        # stem
        stem = tf.keras.layers.Conv1D(filters=16, kernel_size=3, strides=2, padding='valid')(input_y)
        if batchnorm==True:
            stem = tf.keras.layers.BatchNormalization(scale=False)(stem)
        stem = tf.keras.layers.Conv1D(filters=16, kernel_size=3, strides=2, padding='valid')(stem)
        if batchnorm==True:
            stem = tf.keras.layers.BatchNormalization(scale=False)(stem)
        stem = tf.keras.layers.Conv1D(filters=32, kernel_size=3, strides=1, padding='same')(stem)
        if batchnorm==True:
            stem = tf.keras.layers.BatchNormalization(scale=False)(stem)
        stem = tf.keras.layers.MaxPool1D(pool_size=3, strides=2, padding='valid')(stem)
        stem = tf.keras.layers.Conv1D(filters=40, kernel_size=1, strides=1, padding='same')(stem)
        if batchnorm==True:
            stem = tf.keras.layers.BatchNormalization(scale=False)(stem)
        stem = tf.keras.layers.Conv1D(filters=96, kernel_size=3, strides=2, padding='valid')(stem)
        if batchnorm==True:
            stem = tf.keras.layers.BatchNormalization(scale=False)(stem)
        stem = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=2, padding='valid')(stem)

        # A
        IRNA = tf.keras.activations.relu(stem)
        for i in range(5):
            IRNA1 = tf.keras.layers.Conv1D(filters=16, kernel_size=1, strides=1, padding='same')(IRNA)
            if batchnorm==True:
                IRNA1 = tf.keras.layers.BatchNormalization(scale=False)(IRNA1)
            IRNA2 = tf.keras.layers.Conv1D(filters=16, kernel_size=1, strides=1, padding='same')(IRNA)
            if batchnorm==True:
                IRNA2 = tf.keras.layers.BatchNormalization(scale=False)(IRNA2)
            IRNA2 = tf.keras.layers.Conv1D(filters=16, kernel_size=3, strides=1, padding='same')(IRNA2)
            if batchnorm==True:
                IRNA2 = tf.keras.layers.BatchNormalization(scale=False)(IRNA2)
            IRNA3 = tf.keras.layers.Conv1D(filters=16, kernel_size=1, strides=1, padding='same')(IRNA)
            if batchnorm==True:
                IRNA3 = tf.keras.layers.BatchNormalization(scale=False)(IRNA3)
            IRNA3 = tf.keras.layers.Conv1D(filters=16, kernel_size=3, strides=1, padding='same')(IRNA3)
            if batchnorm==True:
                IRNA3 = tf.keras.layers.BatchNormalization(scale=False)(IRNA3)
            IRNA3 = tf.keras.layers.Conv1D(filters=16, kernel_size=3, strides=1, padding='same')(IRNA3)
            if batchnorm==True:
                IRNA3 = tf.keras.layers.BatchNormalization(scale=False)(IRNA3)
            IRNA4 = tf.keras.layers.concatenate([IRNA1,IRNA2,IRNA3],axis=2)
            IRNA4 = tf.keras.layers.Conv1D(filters=128, kernel_size=1, strides=1, padding='same')(IRNA4)
            if batchnorm==True:
                IRNA4 = tf.keras.layers.BatchNormalization(scale=False)(IRNA4)
            IRNA = tf.keras.layers.Add()([IRNA,IRNA4])
            IRNA = tf.keras.activations.relu(IRNA)

        # reduction A
        RA1 = tf.keras.layers.MaxPool1D(pool_size=3, strides=2, padding='valid')(IRNA)
        RA2 = tf.keras.layers.Conv1D(filters=192, kernel_size=3, strides=2, padding='valid')(IRNA)
        if batchnorm==True:
            RA2 = tf.keras.layers.BatchNormalization(scale=False)(RA2)
        RA3 = tf.keras.layers.Conv1D(filters=96, kernel_size=1, strides=1, padding='same')(IRNA)
        if batchnorm==True:
            RA3 = tf.keras.layers.BatchNormalization(scale=False)(RA3)
        RA3 = tf.keras.layers.Conv1D(filters=96, kernel_size=3, strides=1, padding='same')(RA3)
        if batchnorm==True:
            RA3 = tf.keras.layers.BatchNormalization(scale=False)(RA3)
        RA3 = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=2, padding='valid')(RA3)
        if batchnorm==True:
            RA3 = tf.keras.layers.BatchNormalization(scale=False)(RA3)
        RA = tf.keras.layers.concatenate([RA1,RA2,RA3],axis=2)

        flag = 0
        if extra is not None:
            new_input = tf.expand_dims(tf.keras.layers.Dense(RA.shape[1])(extra),axis=-1)
            RA = tf.keras.layers.concatenate([RA,new_input],axis=2)
            flag = 1

        # B
        IRNB = tf.keras.activations.relu(RA)
        for i in range(10):
            IRNB1 = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same')(IRNB)
            if batchnorm==True:
                IRNB1 = tf.keras.layers.BatchNormalization(scale=False)(IRNB1)
            IRNB2 = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same')(IRNB)
            if batchnorm==True:
                IRNB2 = tf.keras.layers.BatchNormalization(scale=False)(IRNB2)
            IRNB2 = tf.keras.layers.Conv1D(filters=64, kernel_size=7, strides=1, padding='same')(IRNB2)
            if batchnorm==True:
                IRNB2 = tf.keras.layers.BatchNormalization(scale=False)(IRNB2)
            IRNB2 = tf.keras.layers.Conv1D(filters=64, kernel_size=1, strides=1, padding='same')(IRNB2)
            if batchnorm==True:
                IRNB2 = tf.keras.layers.BatchNormalization(scale=False)(IRNB2)
            IRNB3 = tf.keras.layers.concatenate([IRNB1,IRNB2],axis=2)
            IRNB3 = tf.keras.layers.Conv1D(filters=448+flag, kernel_size=1, strides=1, padding='same')(IRNB3)
            if batchnorm==True:
                IRNB3 = tf.keras.layers.BatchNormalization(scale=False)(IRNB3)
            IRNB = tf.keras.layers.Add()([IRNB,IRNB3])
            IRNB = tf.keras.activations.relu(IRNB)

        # reduction B
        RB1 = tf.keras.layers.MaxPool1D(pool_size=3, strides=2, padding='valid')(IRNB)
        RB2 = tf.keras.layers.Conv1D(filters=128, kernel_size=1, strides=1, padding='same')(IRNB)
        if batchnorm==True:
            RB2 = tf.keras.layers.BatchNormalization(scale=False)(RB2)
        RB2 = tf.keras.layers.Conv1D(filters=192, kernel_size=3, strides=2, padding='valid')(RB2)
        if batchnorm==True:
            RB2 = tf.keras.layers.BatchNormalization(scale=False)(RB2)
        RB3 = tf.keras.layers.Conv1D(filters=128, kernel_size=1, strides=1, padding='same')(IRNB)
        if batchnorm==True:
            RB3 = tf.keras.layers.BatchNormalization(scale=False)(RB3)
        RB3 = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=2, padding='valid')(RB3)
        if batchnorm==True:
            RB3 = tf.keras.layers.BatchNormalization(scale=False)(RB3)
        RB4 = tf.keras.layers.Conv1D(filters=128, kernel_size=1, strides=1, padding='same')(IRNB)
        if batchnorm==True:
            RB4 = tf.keras.layers.BatchNormalization(scale=False)(RB4)
        RB4 = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=1, padding='same')(RB4)
        if batchnorm==True:
            RB4 = tf.keras.layers.BatchNormalization(scale=False)(RB4)
        RB4 = tf.keras.layers.Conv1D(filters=128, kernel_size=3, strides=2, padding='valid')(RB4)
        if batchnorm==True:
            RB4 = tf.keras.layers.BatchNormalization(scale=False)(RB4)
        RB = tf.keras.layers.concatenate([RB1,RB2,RB3,RB4],axis=2)

        # C
        IRNC = tf.keras.activations.relu(RB)
        for i in range(5):
            IRNC1 = tf.keras.layers.Conv1D(filters=96, kernel_size=1, strides=1, padding='same')(IRNC)
            if batchnorm==True:
                IRNC1 = tf.keras.layers.BatchNormalization(scale=False)(IRNC1)
            IRNC2 = tf.keras.layers.Conv1D(filters=96, kernel_size=1, strides=1, padding='same')(IRNC)
            if batchnorm==True:
                IRNC2 = tf.keras.layers.BatchNormalization(scale=False)(IRNC2)
            IRNC2 = tf.keras.layers.Conv1D(filters=96, kernel_size=3, strides=1, padding='same')(IRNC2)
            if batchnorm==True:
                IRNC2 = tf.keras.layers.BatchNormalization(scale=False)(IRNC2)
            IRNC2 = tf.keras.layers.Conv1D(filters=96, kernel_size=1, strides=1, padding='same')(IRNC2)
            if batchnorm==True:
                IRNC2 = tf.keras.layers.BatchNormalization(scale=False)(IRNC2)
            IRNC3 = tf.keras.layers.concatenate([IRNC1,IRNC2],axis=2)
            IRNC3 = tf.keras.layers.Conv1D(filters=896+flag, kernel_size=1, strides=1, padding='same')(IRNC3)
            if batchnorm==True:
                IRNC3 = tf.keras.layers.BatchNormalization(scale=False)(IRNC3)
            IRNC = tf.keras.layers.Add()([IRNC,IRNC3])
            IRNC = tf.keras.activations.relu(IRNC)

        AP = tf.keras.layers.GlobalAveragePooling1D()(IRNC)
        DO = tf.keras.layers.Dropout(0.2)(AP)
        return DO

    def encode_r1(self, y=None,training=True):
        print(y.shape)
        mean, logvar, weight = tf.split(self.encoder_r1(y,training=training), num_or_size_splits=[self.z_dim*self.n_modes, self.z_dim*self.n_modes,self.n_modes], axis=1)
        mean = tf.reshape(mean,[-1,self.n_modes,self.z_dim])
        logvar = tf.reshape(logvar,[-1,self.n_modes,self.z_dim])
        weight = tf.reshape(weight,[-1,self.n_modes])
        return mean, logvar, weight

    def encode_q(self, x=None, y=None):
        return tf.split(self.encoder_q([y,x]), num_or_size_splits=[self.z_dim,self.z_dim], axis=1)

    def decode_r2(self, y=None, z=None, apply_sigmoid=False):
        return tf.split(self.decoder_r2([y,z]), num_or_size_splits=[self.x_dim_np, 2*self.x_dim_p, 3, self.x_dim_np, self.x_dim_p, 1], axis=1)

    def compute_loss(self, data):

        # unpack data
        x, y = data

        # add noise and randomise distance again
        y = (y + tf.random.normal(shape=tf.shape(y), mean=0.0, stddev=1.0, dtype=tf.float32))/params['y_normscale']
    
        # add correlation channels
        #y_corr = tf.stack([tf.signal.irfft(tf.signal.rfft(y[:,:,0])*tf.signal.rfft(y[:,:,1])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,1])*tf.signal.rfft(y[:,:,2])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,2])*tf.signal.rfft(y[:,:,0]))],axis=2)*self.corr_norm
        #y = tf.keras.layers.concatenate([y,y_corr],axis=2)

        # define r1 encoder
        mean_r1, logvar_r1, logweight_r1 = self.encode_r1(y=y,training=True)
        gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=tf.cast(logweight_r1,dtype=tf.float32)),
            components_distribution=tfd.MultivariateNormalDiag(
            loc=tf.cast(mean_r1,dtype=tf.float32),
            scale_diag=tf.cast(EPS + tf.sqrt(tf.exp(logvar_r1)),dtype=tf.float32)))
    
        # define q encoder
        mean_q, logvar_q = self.encode_q(x=x,y=y)
        mvn_q = tfp.distributions.MultivariateNormalDiag(
            loc=mean_q,
            scale_diag=EPS + tf.sqrt(tf.exp(logvar_q)))
        z_samp = mvn_q.sample()

        # define r2 encoder
        mean_np_r2, mean_p_r2, mean_s_r2, logvar_np_r2, logvar_p_r2, logvar_s_r2 = self.decode_r2(z=z_samp,y=y)

        # truncated normal for non-periodic params
        tmvn_np_r2 = tfp.distributions.TruncatedNormal(
            loc=tf.cast(mean_np_r2,dtype=tf.float32),
            scale=tf.cast(EPS + tf.sqrt(tf.exp(logvar_np_r2)),dtype=tf.float32),
            low=-10.0 + self.ramp*10.0, high=1.0 + 10.0 - self.ramp*10.0)
        tmvn_r2_cost_recon = -1.0*tf.reduce_mean(tf.reduce_sum(tmvn_np_r2.log_prob(tf.boolean_mask(x,nonperiodic_mask,axis=1)),axis=1),axis=0)
    
        # 2D representations of periodic params
        mean_x_r2, mean_y_r2 = tf.split(mean_p_r2, num_or_size_splits=2, axis=1)   # split the means into x and y coords [size (batch,p) each]
        mean_angle_r2 = tf.math.floormod(tf.math.atan2(tf.cast(mean_y_r2,dtype=tf.float32),tf.cast(mean_x_r2,dtype=tf.float32)),2.0*np.pi)  # the mean angle (0,2pi) [size (batch,p)]
        scale_p_r2 = EPS + tf.sqrt(tf.exp(logvar_p_r2))  # define the 2D Gaussian scale [size (batch,2*p)]
        vm_r2 = tfp.distributions.VonMises(
            loc=tf.cast(mean_angle_r2,dtype=tf.float32), #tf.cast(2.0*np.pi*mean_p_r2,dtype=tf.float32), 
            concentration=tf.cast(tf.math.reciprocal(EPS + fourpisq*tf.exp(logvar_p_r2)),dtype=tf.float32)
        )
        vm_r2_cost_recon = -1.0*tf.reduce_mean(tf.reduce_sum(lntwopi + vm_r2.log_prob(2.0*np.pi*tf.boolean_mask(x,periodic_mask,axis=1)),axis=1),axis=0)

        fvm_loc = tf.reshape(tf.math.l2_normalize(mean_s_r2, axis=1),[-1,3])
        fvm_con = tf.reshape(tf.math.reciprocal(EPS + tf.exp(logvar_s_r2)),[-1])
        fvm_r2 = tfp.distributions.VonMisesFisher(
            mean_direction = tf.cast(fvm_loc,dtype=tf.float32), 
            concentration = tf.cast(fvm_con,dtype=tf.float32) 
        )
        ra_sky = tf.reshape(2*np.pi*tf.boolean_mask(x,ra_mask,axis=1),(-1,1))       # convert the scaled 0->1 true RA value back to radians
        dec_sky = tf.reshape(np.pi*(tf.boolean_mask(x,dec_mask,axis=1) - 0.5),(-1,1)) # convert the scaled 0>1 true dec value back to radians
        xyz_unit = tf.concat([tf.cos(ra_sky)*tf.cos(dec_sky),tf.sin(ra_sky)*tf.cos(dec_sky),tf.sin(dec_sky)],axis=1)   # construct the true parameter unit vector
        test = tf.math.l2_normalize(xyz_unit,axis=1)
        test2 = fvm_r2.log_prob(tf.reshape(test,(-1,3)))
        fvm_r2_cost_recon = -1.0*tf.reduce_mean(lnfourpi + fvm_r2.log_prob(test),axis=0)

        simple_cost_recon = tmvn_r2_cost_recon + vm_r2_cost_recon + fvm_r2_cost_recon
    
        selfent_q = -1.0*tf.reduce_mean(mvn_q.entropy())
        log_r1_q = gm_r1.log_prob(tf.cast(z_samp,dtype=tf.float32))   # evaluate the log prob of r1 at the q samples
        cost_KL = tf.cast(selfent_q,dtype=tf.float32) - tf.reduce_mean(log_r1_q)
        return simple_cost_recon, cost_KL

    def call(self, data):
        return None

    def train_step(self, data):
        """Executes one training step and returns the loss.
        This function computes the loss and gradients, and uses the latter to
        update the model's parameters.
        """
        with tf.GradientTape() as tape:
            r_loss, kl_loss = self.compute_loss(data)
            loss = r_loss + self.ramp*kl_loss
            scaled_loss = self.optimizer.get_scaled_loss(loss)
        print('inside train_step ramp = {}'.format(self.ramp))
        scaled_gradients = tape.gradient(scaled_loss, self.trainable_variables)
        gradients = self.optimizer.get_unscaled_gradients(scaled_gradients)
        self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
        return {"r_loss": r_loss, "kl_loss": kl_loss, "loss": loss}

    def test_step(self, data):
        """Executes one training step and returns the loss.
        This function computes the loss and gradients, and uses the latter to
        update the model's parameters.
        """
        r_loss, kl_loss = self.compute_loss(data)
        loss = r_loss + self.ramp*kl_loss
        return {"r_loss": r_loss, "kl_loss": kl_loss, "loss": loss}

    def gen_samples(self, y, nsamples=1000, max_samples=1000):

        y = y/params['y_normscale']
        y = tf.tile(y,(max_samples,1,1))
        samp_iterations = int(nsamples/max_samples)
        
        # add correlation channels
        #y_corr = tf.stack([tf.signal.irfft(tf.signal.rfft(y[:,:,0])*tf.signal.rfft(y[:,:,1])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,1])*tf.signal.rfft(y[:,:,2])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,2])*tf.signal.rfft(y[:,:,0]))],axis=2)*self.corr_norm
        #y = tf.keras.layers.concatenate([y,y_corr],axis=2)

        for i in range(samp_iterations):

            mean_r1, logvar_r1, logweight_r1 = self.encode_r1(y=y)
            scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
            gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=tf.cast(logweight_r1,dtype=tf.float32)),
                components_distribution=tfd.MultivariateNormalDiag(
                loc=tf.cast(mean_r1,dtype=tf.float32),
                scale_diag=tf.cast(scale_r1,dtype=tf.float32)))
            z_samp = gm_r1.sample()
            
            mean_np_r2_temp, mean_p_r2_temp, mean_s_r2_temp, logvar_np_r2_temp, logvar_p_r2_temp, logvar_s_r2_temp = self.decode_r2(z=z_samp,y=y)
            scale_np_r2 = EPS + tf.sqrt(tf.exp(logvar_np_r2_temp))
            
            # truncated Gaussian
            tmvn_np_r2 = tfp.distributions.TruncatedNormal(
                loc=tf.cast(mean_np_r2_temp,dtype=tf.float32),
                scale=tf.cast(scale_np_r2,dtype=tf.float32),
                low=-10.0 + self.ramp*10.0, high=1.0 + 10.0 - self.ramp*10.0)
            mean_x_r2, mean_y_r2 = tf.split(mean_p_r2_temp, num_or_size_splits=2, axis=1)
            mean_angle_r2 = tf.math.floormod(tf.math.atan2(tf.cast(mean_y_r2,dtype=tf.float32),tf.cast(mean_x_r2,dtype=tf.float32)),2.0*np.pi)

            # Von Mises
            vm_r2 = tfp.distributions.VonMises(
                loc=tf.cast(mean_angle_r2,dtype=tf.float32), #tf.cast(2.0*np.pi*mean_p_r2_temp,dtype=tf.float32),
                concentration=tf.cast(tf.math.reciprocal(EPS + fourpisq*tf.exp(logvar_p_r2_temp)),dtype=tf.float32)
            )
        
            # Fisher Von Mises
            fvm_loc = tf.reshape(tf.math.l2_normalize(mean_s_r2_temp, axis=1),[max_samples,3])
            fvm_con = tf.reshape(tf.math.reciprocal(EPS + tf.exp(logvar_s_r2_temp)),[max_samples])
            fvm_r2 = tfp.distributions.VonMisesFisher(
                mean_direction = tf.cast(fvm_loc,dtype=tf.float32),
                concentration = tf.cast(fvm_con,dtype=tf.float32)
            )

            tmvn_x_sample = tmvn_np_r2.sample()
            vm_x_sample = tf.math.floormod(vm_r2.sample(),(2.0*np.pi))/(2.0*np.pi)  # samples are output on range [-pi,pi] so rescale to [0,1]
            xyz = tf.reshape(fvm_r2.sample(),[max_samples,3])          # sample the distribution
            samp_ra = tf.math.floormod(tf.math.atan2(tf.slice(xyz,[0,1],[max_samples,1]),tf.slice(xyz,[0,0],[max_samples,1])),2.0*np.pi)/(2.0*np.pi)   # convert to the rescaled 0->1 RA from the unit vector
            samp_dec = (tf.asin(tf.slice(xyz,[0,2],[max_samples,1])) + 0.5*np.pi)/np.pi                       # convert to the rescaled 0->1 dec from the unit vector
            fvm_x_sample = tf.reshape(tf.concat([samp_ra,samp_dec],axis=1),[max_samples,2])             # group the sky samples        

            if i==0:
                x_sample = tf.gather(tf.concat([tmvn_x_sample,vm_x_sample,fvm_x_sample],axis=1),idx_periodic_mask,axis=1)
                mean_r2 = tf.concat([mean_np_r2_temp,mean_p_r2_temp,mean_s_r2_temp],axis=1)
                logvar_r2 = tf.concat([logvar_np_r2_temp,logvar_p_r2_temp,logvar_s_r2_temp],axis=1)
            else:
                x_sample = tf.concat([x_sample,tf.gather(tf.concat([tmvn_x_sample,vm_x_sample,fvm_x_sample],axis=1),idx_periodic_mask,axis=1)],axis=0)
                mean_r2 = tf.concat([mean_r2,tf.concat([mean_np_r2_temp,mean_p_r2_temp,mean_s_r2_temp],axis=1)],axis=0)
                logvar_r2 = tf.concat([logvar_r2,tf.concat([logvar_np_r2_temp,logvar_p_r2_temp,logvar_s_r2_temp],axis=1)],axis=0)
        
        return x_sample, mean_r2, logvar_r2

    def gen_mode_samples(self, y, nsamples=1000, max_samples=1000):

        y = y/params['y_normscale']
        y = tf.tile(y,(max_samples,1,1))
        samp_iterations = int(nsamples/max_samples)

        # add correlation channels
        #y_corr = tf.stack([tf.signal.irfft(tf.signal.rfft(y[:,:,0])*tf.signal.rfft(y[:,:,1])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,1])*tf.signal.rfft(y[:,:,2])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,2])*tf.signal.rfft(y[:,:,0]))],axis=2)*self.corr_norm
        #y = tf.keras.layers.concatenate([y,y_corr],axis=2)

        x_sample = []
        for k in range(min(4,params['n_modes'])):
            for i in range(samp_iterations):
                mean_r1, logvar_r1, logweight_r1 = self.encode_r1(y=y)
                scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
                idx = np.argsort(logweight_r1[0,:])[::-1]    # get order of weights in reverse order
                weights = np.zeros(params['n_modes'])
                weights[idx[k]] = 1.0   # make the k'th highest mode the only mode   
                gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(probs=tf.cast(weights,dtype=tf.float32)),
                    components_distribution=tfd.MultivariateNormalDiag(
                    loc=tf.cast(mean_r1,dtype=tf.float32),
                    scale_diag=tf.cast(scale_r1,dtype=tf.float32)))
                z_samp = gm_r1.sample()
                mean_np_r2, mean_p_r2, mean_s_r2, logvar_np_r2, logvar_p_r2, logvar_s_r2 = self.decode_r2(z=z_samp,y=y)
                scale_np_r2 = EPS + tf.sqrt(tf.exp(logvar_np_r2))
            
                tmvn_np_r2 = tfp.distributions.TruncatedNormal(
                    loc=tf.cast(mean_np_r2,dtype=tf.float32),
                    scale=tf.cast(scale_np_r2,dtype=tf.float32),
                    low=-10.0 + self.ramp*10.0, high=1.0 + 10.0 - self.ramp*10.0)
                mean_cos_r2, mean_sin_r2 = tf.split(mean_p_r2, num_or_size_splits=2, axis=1)
                mean_angle_r2 = tf.math.floormod(tf.math.atan2(tf.cast(mean_sin_r2,dtype=tf.float32),tf.cast(mean_cos_r2,dtype=tf.float32)),2.0*np.pi)
            
                vm_r2 = tfp.distributions.VonMises(
                    loc=tf.cast(mean_angle_r2,dtype=tf.float32), #tf.cast(2.0*np.pi*mean_p_r2,dtype=tf.float32),
                    concentration=tf.cast(tf.math.reciprocal(EPS + fourpisq*tf.exp(logvar_p_r2)),dtype=tf.float32)
                )
                fvm_loc = tf.reshape(tf.math.l2_normalize(mean_s_r2, axis=1),[max_samples,3])
                fvm_con = tf.reshape(tf.math.reciprocal(EPS + tf.exp(logvar_s_r2)),[max_samples])
                fvm_r2 = tfp.distributions.VonMisesFisher(
                    mean_direction = tf.cast(fvm_loc,dtype=tf.float32),
                    concentration = tf.cast(fvm_con,dtype=tf.float32)
                )
                tmvn_x_sample = tmvn_np_r2.sample()   
                vm_x_sample = tf.math.floormod(vm_r2.sample(),(2.0*np.pi))/(2.0*np.pi)  # samples are output on range [-pi,pi] so rescale to [0,1]
                xyz = tf.reshape(fvm_r2.sample(),[max_samples,3])          # sample the distribution
                samp_ra = tf.math.floormod(tf.math.atan2(tf.slice(xyz,[0,1],[max_samples,1]),tf.slice(xyz,[0,0],[max_samples,1])),2.0*np.pi)/(2.0*np.pi)   # convert to the rescaled 0->1 RA from the unit vector
                samp_dec = (tf.asin(tf.slice(xyz,[0,2],[max_samples,1])) + 0.5*np.pi)/np.pi                       # convert to the rescaled 0->1 dec from the unit vector
                fvm_x_sample = tf.reshape(tf.concat([samp_ra,samp_dec],axis=1),[max_samples,2])             # group the sky samples
                if i==0:
                    temp_x_sample = tf.gather(tf.concat([tmvn_x_sample,vm_x_sample,fvm_x_sample],axis=1),idx_periodic_mask,axis=1)
                else:
                    temp_x_sample = tf.concat([temp_x_sample,tf.gather(tf.concat([tmvn_x_sample,vm_x_sample,fvm_x_sample],axis=1),idx_periodic_mask,axis=1)],axis=0)
            x_sample.append(temp_x_sample)
        return x_sample

    def gen_z_samples(self, x, y, nsamples=1000, max_samples=1000):

        y = y/params['y_normscale']
        y = tf.tile(y,(max_samples,1,1))
        x = tf.tile(x,(max_samples,1))
        samp_iterations = int(nsamples/max_samples)
 
        # add correlation channels
        #y_corr = tf.stack([tf.signal.irfft(tf.signal.rfft(y[:,:,0])*tf.signal.rfft(y[:,:,1])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,1])*tf.signal.rfft(y[:,:,2])),
        #          tf.signal.irfft(tf.signal.rfft(y[:,:,2])*tf.signal.rfft(y[:,:,0]))],axis=2)*self.corr_norm
        #y = tf.keras.layers.concatenate([y,y_corr],axis=2)

        for i in range(samp_iterations):

            mean_r1_temp, logvar_r1, logweight_r1_temp = self.encode_r1(y=y)
            scale_r1 = EPS + tf.sqrt(tf.exp(logvar_r1))
            gm_r1 = tfd.MixtureSameFamily(mixture_distribution=tfd.Categorical(logits=logweight_r1_temp),
                components_distribution=tfd.MultivariateNormalDiag(
                loc=mean_r1_temp,
                scale_diag=scale_r1))
            mean_q_temp, logvar_q = self.encode_q(x=x,y=y)
            scale_q = EPS + tf.sqrt(tf.exp(logvar_q))
            mvn_q = tfp.distributions.MultivariateNormalDiag(
                          loc=mean_q_temp,
                          scale_diag=scale_q)
            if i==0:
                z_samp_q = mvn_q.sample()
                z_samp_r1 = gm_r1.sample()
                logweight_r1 = logweight_r1_temp
                mean_r1 = mean_r1_temp
                mean_q = mean_q_temp
            else:
                z_samp_q = tf.concat([z_samp_q,mvn_q.sample()],axis=0)
                z_samp_r1 = tf.concat([z_samp_r1,gm_r1.sample()],axis=0)
                logweight_r1 = tf.concat([logweight_r1,logweight_r1_temp],axis=0)
                mean_r1 = tf.concat([mean_r1,mean_r1_temp],axis=0)
                mean_q = tf.concat([mean_q,mean_q_temp],axis=0)
    
        return mean_r1, logweight_r1, z_samp_r1, mean_q, z_samp_q

model = CVAE(int(nonperiodic_len), int(periodic_len), params['ndata'], 3, params['z_dimension'], params['n_modes'], run=run)
optimizer = tf.keras.optimizers.Adam(1e-4)
optimizer = mixed_precision.LossScaleOptimizer(optimizer)
model.compile(optimizer, loss=model.compute_loss)
model.fit(training_generator,
          validation_data=validation_generator,
          epochs=50000,
          use_multiprocessing=False,
          workers=4,
          callbacks=[RampCallback(start=100, ramp_length=100, n_cycles=1, ramp_min=0.0, ramp_max=1.0),
                     LossHistory(run=run),
                     ReloadCallback(training_generator, reload_cadence=8),
                     TestCallback(test_generator, pp_generator, bilby_samples, run, n_samples=10000, test_cadence=plot_cadence)])
                     #tf.keras.callbacks.LearningRateScheduler(scheduler)])
                     #tf.keras.callbacks.ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=5000, cooldown=1000, min_lr=1e-6, verbose=1)])
model.save('/data/www.astro/chrism/vitamin_c/%s/model_final' % (run))
exit(0)























